"""Test command-line-interface pytest fixtures."""

from __future__ import annotations

import inspect
import re
import socket
import sys
from contextlib import closing, suppress
from cProfile import Profile
from getpass import getuser
from hashlib import sha1
from os import environ, linesep
from os.path import pathsep, sep
from pathlib import Path
from tempfile import gettempdir
from textwrap import dedent, indent
from typing import TYPE_CHECKING, Any, Literal, cast

from grizzly_cli.utils import rm_rf
from typing_extensions import Self

from test_cli.helpers import run_command

if TYPE_CHECKING:
    from collections.abc import Callable
    from types import TracebackType

    from _pytest.tmpdir import TempPathFactory
    from behave.model import Feature
    from behave.runner import Context

__all__ = [
    'End2EndFixture',
]


BehaveKeyword = Literal['Then', 'Given', 'And', 'When']


class End2EndValidator:
    name: str
    implementation: Any
    table: list[dict[str, str]] | None

    def __init__(
        self,
        name: str,
        implementation: Callable[[Context], None],
        table: list[dict[str, str]] | None = None,
    ) -> None:
        self.name = name
        self.implementation = implementation
        self.table = table

    @property
    def expression(self) -> str:
        lines: list[str] = [f'Then run validator {self.name}_{self.implementation.__name__}']
        if self.table is not None and len(self.table) > 0:
            lines.append(f'  | {" | ".join(list(self.table[0].keys()))} |')
            lines.extend(f'  | {" | ".join(list(row.values()))} |' for row in self.table)

        return '\n'.join(lines)

    @property
    def impl(self) -> str:
        source_lines = inspect.getsource(self.implementation).split('\n')
        source_lines[0] = dedent(source_lines[0].replace('def ', f'def {self.name}_'))
        source = '\n'.join(source_lines)

        return f"""@then(u'run validator {self.name}_{self.implementation.__name__}')
def {self.name}_{self.implementation.__name__}_wrapper(context: Context) -> None:
    {dedent(source)}
    if on_local(context) or on_worker(context):
        {self.name}_{self.implementation.__name__}(context)
"""


class End2EndFixture:
    _tmp_path_factory: TempPathFactory
    _env: dict[str, str]
    _validators: dict[str | None, list[End2EndValidator]]
    _distributed: bool

    _after_features: dict[str, Callable[[Context, Feature], None]]
    _before_features: dict[str, Callable[[Context, Feature], None]]

    _root: Path | None
    _port: int | None = None

    cwd: Path

    profile: Profile | None

    def __init__(self, tmp_path_factory: TempPathFactory, *, distributed: bool) -> None:
        self._tmp_path_factory = tmp_path_factory
        self.cwd = Path.cwd()
        self._env = environ.copy() if sys.platform == 'win32' else {}
        self._validators = {}
        self._root = None
        self._after_features = {}
        self._before_features = {}
        self._distributed = distributed
        self.profile = None

        temp_dir = environ.get('GRIZZLY_TMP_DIR', gettempdir())

        self.log_file = Path(temp_dir) / 'grizzly.log'
        self.log_file.unlink(missing_ok=True)

    @property
    def mode_root(self) -> Path:
        if self._root is None:
            message = 'root is not set'
            raise AttributeError(message)

        if self._distributed:
            return Path('/srv/grizzly')

        return self._root

    @property
    def root(self) -> Path:
        if self._root is None:
            message = 'root is not set'
            raise AttributeError(message)

        return self._root

    @property
    def mode(self) -> str:
        return 'dist' if self._distributed else 'local'

    @property
    def webserver_port(self) -> int:
        if self._port is None:
            self._port = self.find_free_port()

        return self._port

    @property
    def host(self) -> str:
        host = 'master' if self._distributed else '127.0.0.1'

        return f'{host}:{self.webserver_port}'

    def find_free_port(self) -> int:
        with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
            sock.bind(('', 0))
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            return cast('int', sock.getsockname()[1])

    def inject_webserver_module(self, path: Path) -> None:
        assert self._tmp_path_factory._basetemp is not None

        # create webserver module
        webserver_source = self._tmp_path_factory._basetemp.parent / 'tests' / 'test_cli' / 'webserver.py'
        webserver_destination = path / 'features' / 'steps' / 'webserver.py'
        webserver_destination.write_text(webserver_source.read_text())

    def start_webserver_step_impl(self, port: int) -> str:
        return f"""

@then(u'start webserver on master port "{port}"')
def step_start_webserver(context: Context) -> None:
    from grizzly.locust import on_worker
    if on_worker(context):
        return

    import logging
    from importlib.machinery import SourceFileLoader

    logger = logging.getLogger('webserver')

    w = SourceFileLoader(
        'steps.webserver',
        'features/steps/webserver.py',
    ).load_module('steps.webserver')

    webserver = w.Webserver({port})
    logger.info('starting webserver')
    webserver.start(logger)
    logger.info('webserver started')
"""

    def __enter__(self) -> Self:
        if environ.get('PROFILE', None) is not None:
            self.profile = Profile()
            self.profile.enable()
            self._env.update({'PROFILE': 'true'})

        self._root = self._tmp_path_factory.mktemp('test_context')

        virtual_env = environ.get('VIRTUAL_ENV')

        if virtual_env is None or f'{sep}hatch{sep}env{sep}virtual' not in virtual_env:
            virtual_env_path = self.root / 'venv'

            # create virtualenv
            rc, output = run_command(
                [sys.executable, '-m', 'venv', virtual_env_path.name],
                cwd=self.root,
            )

            try:
                assert rc == 0
            except AssertionError:
                print(''.join(output))

                with self.log_file.open('a+') as fd:
                    fd.write(''.join(output))

                raise
        else:
            virtual_env_path = Path(virtual_env)

        path = environ.get('PATH', '')

        virtual_env_bin_dir = 'Scripts' if sys.platform == 'win32' else 'bin'

        self._env.update(
            {
                'PATH': f'{virtual_env_path!s}{sep}{virtual_env_bin_dir}{pathsep}{path}',
                'VIRTUAL_ENV': f'{virtual_env_path!s}',
                'PYTHONPATH': environ.get('PYTHONPATH', '.'),
                'HOME': environ.get('HOME', '/'),
            }
        )

        if sys.platform == 'win32':
            self._env.update(
                {
                    'SYSTEMROOT': environ['SYSTEMROOT'],
                    'SYSTEMDRIVE': environ['SYSTEMDRIVE'],
                    'USERPROFILE': environ['USERPROFILE'],
                    'PYTHONIOENCODING': 'utf-8',
                    'PYTHONUTF8': '1',
                }
            )

        for env_key in ['SSH_AUTH_SOCK', 'GRIZZLY_MOUNT_CONTEXT']:
            env_value = environ.get(env_key, None)
            if env_value is not None:
                self._env.update({env_key: env_value})

        repo_root_path = (Path(__file__).parent / '..' / '..' / '..').resolve()
        command = ['uv', 'sync', '--active', '--locked', '--package', 'grizzly-loadtester']
        rc, output = run_command(
            command,
            cwd=repo_root_path,
            env=self._env,
        )

        try:
            assert rc == 0
        except AssertionError:
            print(''.join(output))
            with self.log_file.open('a+') as fd:
                fd.write(''.join(output))
            raise

        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        traceback: TracebackType | None,
    ) -> Literal[True]:
        if exc is None:
            if environ.get('KEEP_FILES', None) is None:
                with suppress(AttributeError):
                    rm_rf(self.root)
            else:
                print(self._root)

        if self.profile is not None:
            self.profile.disable()
            self.profile.dump_stats('grizzly-cli-e2e-tests.hprof')

        return True

    def add_validator(
        self,
        implementation: Callable[[Context], None],
        scenario: str | None = None,
        table: list[dict[str, str]] | None = None,
    ) -> None:
        callee = inspect.stack()[1].function

        if self._validators.get(scenario, None) is None:
            self._validators[scenario] = []

        self._validators[scenario].append(End2EndValidator(callee, implementation, table))

    def add_after_feature(self, implementation: Callable[[Context, Feature], None]) -> None:
        callee = inspect.stack()[1].function
        self._after_features[callee] = implementation

    def add_before_feature(self, implementation: Callable[[Context, Feature], None]) -> None:
        callee = inspect.stack()[1].function
        self._before_features[callee] = implementation

    def modify_feature(self, feature_lines: list[str]) -> list[str]:
        scenario: str | None = None
        indentation = '    '
        modified_feature_lines: list[str] = []
        offset = 0  # number of added steps

        for nr, line in enumerate(feature_lines):
            modified_feature_lines.append(line)

            last_line = nr == len(feature_lines) - 1
            scenario_definition = line.strip().startswith('Scenario:')

            if scenario_definition or last_line:
                if scenario is not None:
                    validators = self._validators.get(scenario, self._validators.get(None, None))
                    if validators is not None:
                        for validator in validators:
                            validator_expression = indent(f'{validator.expression}', prefix=indentation * 2)
                            index = nr + offset
                            while modified_feature_lines[index].strip() == '' or 'Scenario:' in modified_feature_lines[index]:
                                index -= 1

                            index += 1
                            modified_feature_lines.insert(index, validator_expression)

                            offset += 1

                if scenario_definition:
                    scenario = line.replace('Scenario:', '').strip()
                    indentation, _ = line.split('Scenario:', 1)

        modified_feature_lines.append('')

        return modified_feature_lines

    def create_feature(self, contents: str, name: str | None = None, identifier: str | None = None) -> str:
        if name is None:
            name = inspect.stack()[1].function

        if identifier is not None:
            identifier = sha1(identifier.encode()).hexdigest()[:8]  # noqa: S324
            name = f'{name}_{identifier}'

        feature_lines = contents.strip().split('\n')
        feature_lines[0] = f'Feature: {name}'
        steps_file = self.root / 'features' / 'steps' / 'steps.py'
        environment_file = self.root / 'features' / 'environment.py'

        modified_feature_lines = self.modify_feature(feature_lines)

        contents = '\n'.join(modified_feature_lines)

        # write feature file
        with (self.root / 'features' / f'{name}.feature').open('w+') as fd:
            fd.write(contents)

        feature_file_name = fd.name.replace(f'{self.root}/', '')

        # cache current step implementations
        with steps_file.open('r') as fd:
            steps_impl = fd.read()

        # add step implementations
        with steps_file.open('a') as fd:
            added_validators: list[str] = []
            for validators in self._validators.values():
                for validator in validators:
                    # write expression and step implementation to steps/steps.py
                    if validator.impl not in steps_impl and validator.impl not in added_validators:
                        fd.write(f'\n\n{validator.impl}')
                        added_validators.append(validator.impl)

            added_validators = []

        # add after_feature hook, always write all of 'em
        with environment_file.open('w') as fd:
            fd.write('from typing import Any, Tuple, Dict, cast\n\n')
            fd.write('from behave.runner import Context\n')
            fd.write('from behave.model import Feature\n')
            fd.write('from grizzly.context import GrizzlyContext\n')
            fd.write(
                'from grizzly.environment import before_feature as grizzly_before_feature, '
                'after_feature as grizzly_after_feature, before_scenario, after_scenario, before_step\n\n',
            )

            fd.write('def before_feature(context: Context, feature: Feature, *args: Tuple[Any, ...], **kwargs: Dict[str, Any]) -> None:\n')
            if len(self._before_features) > 0:
                for feature_name in self._before_features:
                    fd.write(f'    if feature.name == "{feature_name}":\n')
                    fd.write(f'        {feature_name}_before_feature(context, feature)\n\n')
            fd.write('    grizzly_before_feature(context, feature)\n\n')

            for key, before_feature_impl in self._before_features.items():
                source_lines = dedent(inspect.getsource(before_feature_impl)).split('\n')
                source_lines[0] = re.sub(r'^def .*?\(', f'def {key}_before_feature(', source_lines[0])
                source = '\n'.join(source_lines)

                fd.write(source + '\n\n')

            fd.write('def after_feature(context: Context, feature: Feature, *args: Tuple[Any, ...], **kwargs: Dict[str, Any]) -> None:\n')
            fd.write('    grizzly_after_feature(context, feature)\n\n')
            if len(self._after_features) > 0:
                for feature_name in self._after_features:
                    fd.write(f'    if feature.name == "{feature_name}":\n')
                    fd.write(f'        {feature_name}_after_feature(context, feature)\n\n')

            for key, after_feature_impl in self._after_features.items():
                source_lines = dedent(inspect.getsource(after_feature_impl)).split('\n')
                source_lines[0] = re.sub(r'^def .*?\(', f'def {key}_after_feature(', source_lines[0])
                source = '\n'.join(source_lines)

                fd.write(source + '\n\n')

        # step validators are are now "burned"...
        self._validators.clear()

        return feature_file_name

    def execute(
        self,
        feature_file: Path,
        env_conf_file: str | None = None,
        testdata: dict[str, str] | None = None,
        cwd: Path | None = None,
        arguments: list[str] | None = None,
    ) -> tuple[int, list[str]]:
        if arguments is None:
            arguments = []

        command: list[str] = [
            'grizzly-cli',
            self.mode,
            'run',
            *arguments,
            '--yes',
            '--verbose',
            feature_file.as_posix(),
        ]

        if self._distributed:
            command = [*command[:2], '--project-name', self.root.name, *command[2:]]

        if env_conf_file is not None:
            command += ['-e', env_conf_file]

        if testdata is not None:
            for key, value in testdata.items():
                command += ['-T', f'{key}={value}']

        rc, output = run_command(
            command,
            cwd=cwd or self.mode_root,
            env=self._env,
        )

        if sys.platform == 'win32':
            output = [line.replace(linesep, '\n') for line in output]

        if rc != 0:
            print(''.join(output))

            with self.log_file.open('a+') as fd:
                fd.write(''.join(output))

                for container in ['master', 'worker'] if self._distributed else []:
                    command = ['docker', 'container', 'logs', f'{self.root.name}-{getuser()}_{container}_1']
                    _, output = run_command(
                        command,
                        cwd=self.root,
                        env=self._env,
                    )

                    print(''.join(output))
                    fd.write(''.join(output))

        return rc, output



"""Base grizzly-cli module."""

from __future__ import annotations

from importlib.metadata import PackageNotFoundError, version
from os import environ
from pathlib import Path
from typing import TYPE_CHECKING, ClassVar

try:
    __version__ = version('grizzly-loadtester-cli')
except PackageNotFoundError:
    __version__ = 'unkown'

try:
    __common_version__ = version('grizzly-loadtester-common')
except PackageNotFoundError:
    __common_version__ = 'unknown'

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Callable

    from behave.model import Scenario

    from grizzly_cli.argparse import ArgumentSubParser

EXECUTION_CONTEXT = Path.cwd().as_posix()

STATIC_CONTEXT = Path.joinpath(Path(__file__).parent.absolute(), 'static').as_posix()

MOUNT_CONTEXT = environ.get('GRIZZLY_MOUNT_CONTEXT', EXECUTION_CONTEXT)

PROJECT_NAME = Path(EXECUTION_CONTEXT).name

SCENARIOS: list[Scenario] = []

FEATURE_DESCRIPTION: str | None = None


class register_parser:
    registered: ClassVar[list[Callable[[ArgumentSubParser], None]]] = []
    order: int | None

    def __init__(self, order: int | None = None) -> None:
        self.order = order

    def __call__(self, func: Callable[[ArgumentSubParser], None]) -> Callable[[ArgumentSubParser], None]:
        if self.order is not None:
            self.registered.insert(self.order - 1, func)
        else:
            self.registered.append(func)

        return func


__all__ = ['__common_version__', '__version__']



"""Test command-line-interface helpers."""

from __future__ import annotations

import os
import subprocess
import sys
from abc import ABCMeta
from contextlib import contextmanager, suppress
from importlib.metadata import version
from pathlib import Path
from typing import TYPE_CHECKING, Any

from behave.model import Scenario, Step
from grizzly_cli.utils import rm_rf

if TYPE_CHECKING:
    from collections.abc import Generator

__all__ = ['rm_rf']


def CaseInsensitive(value: str) -> object:  # noqa: N802
    class Wrapped(str):
        __slots__ = ()

        def __eq__(self, other: object) -> bool:
            return isinstance(other, str) and other.lower() == value.lower()

        def __ne__(self, other: object) -> bool:
            return not self.__eq__(other)

        def __neq__(self, other: object) -> bool:
            return self.__ne__(other)

        def __hash__(self) -> int:
            return hash(self)

    return Wrapped()


def run_command(command: list[str], env: dict[str, str] | None = None, cwd: Path | None = None, stdin: str | None = None) -> tuple[int, list[str]]:
    output: list[str] = []
    if env is None:
        env = os.environ.copy()

    if cwd is None:
        cwd = Path.cwd()

    process = subprocess.Popen(
        command,
        env=env,
        cwd=cwd,
        stderr=subprocess.STDOUT,
        stdout=subprocess.PIPE,
        stdin=subprocess.PIPE,
    )

    if stdin is not None:
        assert process.stdin is not None
        process.stdin.write(f'{stdin}\n'.encode())
        process.stdin.close()

    try:
        while process.poll() is None:
            stdout = process.stdout
            if stdout is None:
                break

            buffer = stdout.readline()
            if not buffer:
                break

            line = buffer.decode('utf-8')
            if sys.platform == 'win32':
                line = line.replace(os.linesep, '\n')

            output.append(line)

        process.terminate()
    except KeyboardInterrupt:
        pass
    finally:
        with suppress(Exception):
            process.kill()

    process.wait()

    return process.returncode, output


def create_scenario(name: str, background_steps: list[str], steps: list[str]) -> Scenario:
    scenario = Scenario('', '', '', name)

    for background_step in background_steps:
        [keyword, name] = background_step.split(' ', 1)
        behave_step = Step('', '', keyword.strip(), keyword.strip(), name.strip())
        if scenario._background_steps is None:
            scenario._background_steps = []
        scenario._background_steps.append(behave_step)

    for step in steps:
        [keyword, name] = step.split(' ', 1)
        behave_step = Step('', '', keyword.strip(), keyword.strip(), name.strip())
        scenario.steps.append(behave_step)

    return scenario


def get_current_version() -> tuple[str, str]:
    return version('grizzly-loadtester-cli'), version('grizzly-loadtester-common')


@contextmanager
def cwd(path: Path) -> Generator[None, None, None]:
    current_cwd = Path.cwd()
    os.chdir(path)

    try:
        yield
    finally:
        os.chdir(current_cwd)


def ANY(*cls: type, message: str | None = None) -> object:  # noqa: N802
    """Compare equal to everything, as long as it is of the same type."""

    class WrappedAny(metaclass=ABCMeta):  # noqa: B024
        def __eq__(self, other: object) -> bool:
            if len(cls) < 1:
                return True

            return isinstance(other, cls) and (message is None or (message is not None and message in str(other)))

        def __ne__(self, other: object) -> bool:
            return not self.__eq__(other)

        def __neq__(self, other: object) -> bool:
            return self.__ne__(other)

        def __repr__(self) -> str:
            c = cls[0] if len(cls) == 1 else cls
            representation: list[str] = [f'<ANY({c})', '>']

            if message is not None:
                representation.insert(-1, f", message='{message}'")

            return ''.join(representation)

        def __hash__(self) -> int:
            return hash(self)

    for c in cls:
        WrappedAny.register(c)

    return WrappedAny()


def SOME(cls: type, *value: Any, **values: Any) -> object:  # noqa: N802
    class WrappedSome:
        def __eq__(self, other: object) -> bool:
            if issubclass(cls, dict):

                def get_value(other: Any, attr: str) -> Any:
                    return other.get(attr)
            else:

                def get_value(other: Any, attr: str) -> Any:
                    return getattr(other, attr)

            return isinstance(other, cls) and all(get_value(other, attr) == value for attr, value in values.items())

        def __ne__(self, other: object) -> bool:
            return not self.__eq__(other)

        def __neq__(self, other: object) -> bool:
            return self.__ne__(other)

        def __repr__(self) -> str:
            info = ', '.join([f'{key}={value}' for key, value in values.items()])
            return f'<SOME({cls}, {info})>'

        def __hash__(self) -> int:
            return hash(self)

    if len(value) > 0 and len(values) > 0:
        message = 'cannot use both positional and named arguments'
        raise RuntimeError(message)

    if len(values) < 1 and len(value) < 1:
        raise AttributeError(name='values', obj=str(type))

    if len(value) > 1:
        message = 'can only use 1 positional argument'
        raise RuntimeError(message)

    if len(value) > 0 and isinstance(value[0], dict):
        values = {**value[0]}

    return WrappedSome()



"""Tests for grizzly_cli.run."""

from __future__ import annotations

import logging
from argparse import ArgumentParser
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING

import pytest
from grizzly_cli.run import create_parser, run
from grizzly_cli.utils import setup_logging
from grizzly_cli.utils.configuration import ScenarioTag
from jinja2 import Environment

from test_cli.helpers import CaseInsensitive, rm_rf

if TYPE_CHECKING:
    from _pytest.capture import CaptureFixture
    from _pytest.logging import LogCaptureFixture
    from _pytest.tmpdir import TempPathFactory
    from pytest_mock import MockerFixture


def test_run(capsys: CaptureFixture, mocker: MockerFixture, tmp_path_factory: TempPathFactory) -> None:  # noqa: PLR0915
    setup_logging()

    test_context = tmp_path_factory.mktemp('test_context')
    execution_context = test_context / 'execution-context'
    execution_context.mkdir()
    mount_context = test_context / 'mount-context'
    mount_context.mkdir()
    feature_file = execution_context / 'features' / 'test.feature'
    feature_file.parent.mkdir(parents=True, exist_ok=True)
    feature_file.write_text('Feature: this feature is testing something')
    (execution_context / 'configuration.yaml').write_text('configuration:')

    parser = ArgumentParser()

    sub_parsers = parser.add_subparsers(dest='test')

    create_parser(sub_parsers, parent='local')

    try:
        mocker.patch('grizzly_cli.run.grizzly_cli.EXECUTION_CONTEXT', execution_context.as_posix())
        mocker.patch('grizzly_cli.run.grizzly_cli.MOUNT_CONTEXT', mount_context.as_posix())
        mocker.patch('grizzly_cli.run.get_hostname', return_value='localhost')
        mocker.patch('grizzly_cli.run.find_variable_names_in_questions', side_effect=[['foo', 'bar'], [], [], [], [], [], [], []])
        mocker.patch('grizzly_cli.run.find_metadata_notices', side_effect=[[], ['is the event log cleared?'], ['hello world', 'foo bar'], [], [], [], [], []])
        mocker.patch('grizzly_cli.run.distribution_of_users_per_scenario', autospec=True)
        ask_yes_no_mock = mocker.patch('grizzly_cli.run.ask_yes_no', autospec=True)
        distributed_mock = mocker.MagicMock(return_value=0)
        local_mock = mocker.MagicMock(return_value=0)
        get_input_mock = mocker.patch('grizzly_cli.run.get_input', side_effect=['bar', 'foo'])

        setattr(getattr(run, '__wrapped__'), '__value__', execution_context.as_posix())  # noqa: B009, B010

        arguments = parser.parse_args(
            [
                'run',
                '-e',
                f'{execution_context.as_posix()}/configuration.yaml',
                f'{execution_context.as_posix()}/features/test.feature',
                '--verbose',
            ]
        )
        arguments.file = ' '.join(arguments.file)

        assert run(arguments, distributed_mock) == 0

        capture = capsys.readouterr()
        assert capture.out == ''
        assert (
            capture.err
            == """feature file requires values for 2 variables
the following values was provided:
foo = bar
bar = foo
"""
        )

        local_mock.assert_not_called()
        distributed_mock.assert_called_once_with(
            arguments,
            {
                'GRIZZLY_CLI_HOST': 'localhost',
                'GRIZZLY_EXECUTION_CONTEXT': execution_context.as_posix(),
                'GRIZZLY_MOUNT_CONTEXT': mount_context.as_posix(),
                'GRIZZLY_CONFIGURATION_FILE': CaseInsensitive(Path.joinpath(execution_context, 'configuration.lock.yaml').as_posix()),
                'TESTDATA_VARIABLE_foo': 'bar',
                'TESTDATA_VARIABLE_bar': 'foo',
            },
            {
                'master': [],
                'worker': [],
                'common': ['--verbose', '--no-logcapture', '--no-capture', '--no-capture-stderr'],
            },
        )
        distributed_mock.reset_mock()

        ask_yes_no_mock.assert_called_once_with('continue?')
        ask_yes_no_mock.reset_mock()
        assert get_input_mock.call_count == 2
        args, kwargs = get_input_mock.call_args_list[0]
        assert kwargs == {}
        assert args[0] == 'initial value for "foo": '
        args, kwargs = get_input_mock.call_args_list[1]
        assert kwargs == {}
        assert args[0] == 'initial value for "bar": '
        get_input_mock.reset_mock()

        assert capture.out == ''
        assert capture.err == ('feature file requires values for 2 variables\nthe following values was provided:\nfoo = bar\nbar = foo\n')

        arguments = parser.parse_args(
            [
                'run',
                '-e',
                f'{execution_context.as_posix()}/configuration.yaml',
                f'{execution_context.as_posix()}/features/test.feature',
            ]
        )
        arguments.file = ' '.join(arguments.file)

        assert run(arguments, local_mock) == 0

        capture = capsys.readouterr()

        distributed_mock.assert_not_called()
        local_mock.assert_called_once_with(
            arguments,
            {
                'GRIZZLY_CLI_HOST': 'localhost',
                'GRIZZLY_EXECUTION_CONTEXT': execution_context.as_posix(),
                'GRIZZLY_MOUNT_CONTEXT': mount_context.as_posix(),
                'GRIZZLY_CONFIGURATION_FILE': CaseInsensitive(Path.joinpath(execution_context, 'configuration.lock.yaml').as_posix()),
            },
            {
                'master': [],
                'worker': [],
                'common': [],
            },
        )
        local_mock.reset_mock()

        ask_yes_no_mock.assert_called_once_with('is the event log cleared?')
        ask_yes_no_mock.reset_mock()
        get_input_mock.assert_not_called()

        assert capture.err == ''
        assert capture.out == ''

        # with --yes, notices should only be printed, and not needed to be confirmed via ask_yes_no
        arguments = parser.parse_args(
            [
                'run',
                '-e',
                f'{execution_context.as_posix()}/configuration.yaml',
                '--yes',
                f'{execution_context.as_posix()}/features/test.feature',
            ]
        )
        arguments.file = ' '.join(arguments.file)

        assert run(arguments, local_mock) == 0

        distributed_mock.assert_not_called()

        local_mock.assert_called_once_with(
            arguments,
            {
                'GRIZZLY_CLI_HOST': 'localhost',
                'GRIZZLY_EXECUTION_CONTEXT': execution_context.as_posix(),
                'GRIZZLY_MOUNT_CONTEXT': mount_context.as_posix(),
                'GRIZZLY_CONFIGURATION_FILE': CaseInsensitive(Path.joinpath(execution_context, 'configuration.lock.yaml').as_posix()),
            },
            {
                'master': [],
                'worker': [],
                'common': [],
            },
        )
        local_mock.reset_mock()

        ask_yes_no_mock.assert_not_called()
        get_input_mock.assert_not_called()

        assert capture.err == ''
        assert capture.out == ''

        # no `csv_prefix` nothing should be added
        arguments = parser.parse_args(
            [
                'run',
                '-e',
                f'{execution_context.as_posix()}/configuration.yaml',
                '--yes',
                f'{execution_context.as_posix()}/features/test.feature',
                '--csv-interval',
                '20',
            ]
        )
        arguments.file = ' '.join(arguments.file)

        assert run(arguments, local_mock) == 0

        local_mock.assert_called_once_with(
            arguments,
            {
                'GRIZZLY_CLI_HOST': 'localhost',
                'GRIZZLY_EXECUTION_CONTEXT': execution_context.as_posix(),
                'GRIZZLY_MOUNT_CONTEXT': mount_context.as_posix(),
                'GRIZZLY_CONFIGURATION_FILE': CaseInsensitive(Path.joinpath(execution_context, 'configuration.lock.yaml').as_posix()),
            },
            {
                'master': [],
                'worker': [],
                'common': [],
            },
        )
        local_mock.reset_mock()
        distributed_mock.assert_not_called()

        # static csv-prefix
        arguments = parser.parse_args(
            [
                'run',
                '-e',
                f'{execution_context.as_posix()}/configuration.yaml',
                '--yes',
                f'{execution_context.as_posix()}/features/test.feature',
                '--csv-interval',
                '20',
                '--csv-prefix',
                'test test',
            ]
        )
        arguments.file = ' '.join(arguments.file)

        assert run(arguments, local_mock) == 0

        distributed_mock.assert_not_called()
        local_mock.assert_called_once_with(
            arguments,
            {
                'GRIZZLY_CLI_HOST': 'localhost',
                'GRIZZLY_EXECUTION_CONTEXT': execution_context.as_posix(),
                'GRIZZLY_MOUNT_CONTEXT': mount_context.as_posix(),
                'GRIZZLY_CONFIGURATION_FILE': CaseInsensitive(Path.joinpath(execution_context, 'configuration.lock.yaml').as_posix()),
            },
            {
                'master': [],
                'worker': [],
                'common': ['-Dcsv-prefix="test test"', '-Dcsv-interval=20'],
            },
        )
        local_mock.reset_mock()

        # dynamic csv-prefix
        datetime_mock = mocker.patch(
            'grizzly_cli.run.datetime',
            side_effect=lambda *args, **kwargs: datetime(*args, **kwargs),  # noqa: DTZ001
        )
        datetime_mock.now.return_value = datetime(2022, 12, 6, 13, 1, 13)  # noqa: DTZ001
        arguments = parser.parse_args(
            [
                'run',
                '-e',
                f'{execution_context.as_posix()}/configuration.yaml',
                '--yes',
                f'{execution_context.as_posix()}/features/test.feature',
                '--csv-prefix',
                '--csv-interval',
                '20',
                '--csv-flush-interval',
                '60',
            ]
        )
        arguments.file = ' '.join(arguments.file)

        assert run(arguments, distributed_mock) == 0

        local_mock.assert_not_called()
        distributed_mock.assert_called_once_with(
            arguments,
            {
                'GRIZZLY_CLI_HOST': 'localhost',
                'GRIZZLY_EXECUTION_CONTEXT': execution_context.as_posix(),
                'GRIZZLY_MOUNT_CONTEXT': mount_context.as_posix(),
                'GRIZZLY_CONFIGURATION_FILE': CaseInsensitive(Path.joinpath(execution_context, 'configuration.lock.yaml').as_posix()),
            },
            {
                'master': [],
                'worker': [],
                'common': ['-Dcsv-prefix="this_feature_is_testing_something_20221206T130113"', '-Dcsv-interval=20', '-Dcsv-flush-interval=60'],
            },
        )
        distributed_mock.reset_mock()

        arguments.csv_prefix = None
        arguments.csv_flush_interval = None

        # --log-dir
        arguments = parser.parse_args(
            [
                'run',
                '-e',
                f'{execution_context.as_posix()}/configuration.yaml',
                '--yes',
                '--log-dir',
                'foobar',
                f'{execution_context.as_posix()}/features/test.feature',
            ]
        )
        arguments.file = ' '.join(arguments.file)

        assert run(arguments, distributed_mock) == 0

        local_mock.assert_not_called()
        distributed_mock.assert_called_once_with(
            arguments,
            {
                'GRIZZLY_CLI_HOST': 'localhost',
                'GRIZZLY_EXECUTION_CONTEXT': execution_context.as_posix(),
                'GRIZZLY_MOUNT_CONTEXT': mount_context.as_posix(),
                'GRIZZLY_CONFIGURATION_FILE': CaseInsensitive(Path.joinpath(execution_context, 'configuration.lock.yaml').as_posix()),
                'GRIZZLY_LOG_DIR': 'foobar',
            },
            {
                'master': [],
                'worker': [],
                'common': [],
            },
        )
        distributed_mock.reset_mock()

        capsys.readouterr()

        # --dry-run
        arguments = parser.parse_args(
            [
                'run',
                '-e',
                f'{execution_context.as_posix()}/configuration.yaml',
                '--yes',
                '--log-dir',
                'foobar',
                f'{execution_context.as_posix()}/features/test.feature',
                '--dry-run',
            ]
        )
        arguments.file = ' '.join(arguments.file)

        assert run(arguments, distributed_mock) == 0

        local_mock.assert_not_called()
        distributed_mock.assert_called_once_with(
            arguments,
            {
                'GRIZZLY_CLI_HOST': 'localhost',
                'GRIZZLY_EXECUTION_CONTEXT': execution_context.as_posix(),
                'GRIZZLY_MOUNT_CONTEXT': mount_context.as_posix(),
                'GRIZZLY_CONFIGURATION_FILE': CaseInsensitive(Path.joinpath(execution_context, 'configuration.lock.yaml').as_posix()),
                'GRIZZLY_LOG_DIR': 'foobar',
                'GRIZZLY_DRY_RUN': 'true',
            },
            {
                'master': [],
                'worker': [],
                'common': [],
            },
        )
        distributed_mock.reset_mock()

        capsys.readouterr()

        # --dump
        arguments = parser.parse_args(
            [
                'run',
                '-e',
                f'{execution_context.as_posix()}/configuration.yaml',
                '--yes',
                f'{execution_context.as_posix()}/features/test.feature',
                '--dump',
            ]
        )
        arguments.file = ' '.join(arguments.file)

        assert run(arguments, distributed_mock) == 0

        distributed_mock.assert_not_called()
        local_mock.assert_not_called()

        capture = capsys.readouterr()

        assert capture.out == 'Feature: this feature is testing something\n'
        assert capture.err == ''
    finally:
        rm_rf(test_context)


def test_run_dump(capsys: CaptureFixture, mocker: MockerFixture, tmp_path_factory: TempPathFactory) -> None:  # noqa: PLR0915
    setup_logging()

    test_context = tmp_path_factory.mktemp('test_context')
    execution_context = test_context / 'execution-context'
    execution_context.mkdir()
    mount_context = test_context / 'mount-context'
    mount_context.mkdir()
    feature_file = execution_context / 'features' / 'test.feature'
    feature_file.parent.mkdir(parents=True, exist_ok=True)
    feature_file.write_text('Feature: this feature is testing something')
    (execution_context / 'configuration.yaml').write_text('configuration:')

    parser = ArgumentParser()

    sub_parsers = parser.add_subparsers(dest='test')

    create_parser(sub_parsers, parent='local')

    try:
        mocker.patch('grizzly_cli.run.grizzly_cli.EXECUTION_CONTEXT', execution_context.as_posix())
        mocker.patch('grizzly_cli.run.grizzly_cli.MOUNT_CONTEXT', mount_context.as_posix())
        mocker.patch('grizzly_cli.run.get_hostname', return_value='localhost')
        mocker.patch('grizzly_cli.run.find_variable_names_in_questions', side_effect=[['foo', 'bar'], [], [], [], [], [], [], []])
        mocker.patch('grizzly_cli.run.find_metadata_notices', side_effect=[[], ['is the event log cleared?'], ['hello world', 'foo bar'], [], [], [], [], []])
        mocker.patch('grizzly_cli.run.distribution_of_users_per_scenario', autospec=True)
        distributed_mock = mocker.MagicMock(return_value=0)
        local_mock = mocker.MagicMock(return_value=0)

        setattr(getattr(run, '__wrapped__'), '__value__', execution_context.as_posix())  # noqa: B009, B010

        # --dump output.feature
        feature_file.write_text("""Feature: a feature
    Background: common
        Given a common step

    Scenario: first
        Given a variable with value "{{foo * 0.25 | int }}" and another value " {{ bar |int + 12}}"
        And a variable with value "{{ hello }}"
        And a variable with value "{{ thisshouldwork | upper }}"
        And a variable with value "{{thisshouldalsowork |bigtime}}"
        And a variable with value "{{andthis|too}}"

    Scenario: second
        {% scenario "second", feature="../second.feature" %}

    Scenario: third
        Given a variable with value "{{ some*0.25 | more}}" and another value "{{yes|box }}"

    Scenario: fourth
        {% scenario "fourth", feature="./fourth.feature", foo="bar" %}
""")
        feature_file_2 = execution_context / 'second.feature'
        feature_file_2.write_text("""Feature: a second feature
    Background: common
        Given a common step

    Scenario: second
        Given a variable with value "{{ foobar }}"
        Then run a bloody test
""")
        feature_file_3 = execution_context / 'features' / 'fourth.feature'
        feature_file_3.write_text("""Feature: a fourth feature
    Background: common
        Given a common step

    Scenario: fourth
        Given a variable with value "{{ {$ foo $}_barfoo }}"
        Then get "{{ bar{$ foo $}foo }}" from "{{ {$ foo $}_barfoo }}"
        Then run a bloody test
""")

        arguments = parser.parse_args(
            [
                'run',
                '-e',
                f'{execution_context}/configuration.yaml',
                '--yes',
                f'{execution_context}/features/test.feature',
                '--dump',
                f'{execution_context}/output.feature',
            ]
        )
        arguments.file = ' '.join(arguments.file)

        assert run(arguments, local_mock) == 0

        distributed_mock.assert_not_called()
        local_mock.assert_not_called()

        capture = capsys.readouterr()
        assert capture.err == ''
        assert capture.out == ''

        output_file = execution_context / 'output.feature'
        assert (
            output_file.read_text()
            == """Feature: a feature
    Background: common
        Given a common step

    Scenario: first
        Given a variable with value "{{foo * 0.25 | int }}" and another value " {{ bar |int + 12}}"
        And a variable with value "{{ hello }}"
        And a variable with value "{{ thisshouldwork | upper }}"
        And a variable with value "{{thisshouldalsowork |bigtime}}"
        And a variable with value "{{andthis|too}}"

    Scenario: second
        Given a variable with value "{{ foobar }}"
        Then run a bloody test

    Scenario: third
        Given a variable with value "{{ some*0.25 | more}}" and another value "{{yes|box }}"

    Scenario: fourth
        Given a variable with value "{{ bar_barfoo }}"
        Then get "{{ barbarfoo }}" from "{{ bar_barfoo }}"
        Then run a bloody test
"""
        )

        feature_file.write_text("""Feature: a feature
    Scenario: first
        Given a variable with value "{{ hello }}"

    Scenario: second
        {% scenario "second", feature="../second.feature", prefix="s1" %}

    # Scenario: third
    #     {% scenario "inactive-second", feature="./second.feature", prefix="s1" %}

    Scenario: third
        {% scenario "third", feature="../second.feature", prefix="s1" %}
""")
        feature_file_2 = execution_context / 'second.feature'
        feature_file_2.write_text("""Feature: a second feature
    Background: common
        Given a common step

    Scenario: second
        Given a variable with value "{{ {$ prefix $}foobar }}"
        Then run a bloody test
            \"\"\"
            with step text
            that spans
            more than
            one line
            \"\"\"

    Scenario: third
        Given a variable with value "{{ {$ prefix $}value }}"
        Then run a bloody test, with table
          | hello | world |
          | foo   | bar   |
          | bar   |       |
          |       | foo   |
""")

        output_file.unlink(missing_ok=True)

        arguments = parser.parse_args(
            [
                'run',
                '-e',
                f'{execution_context.as_posix()}/configuration.yaml',
                '--yes',
                f'{execution_context.as_posix()}/features/test.feature',
                '--dump',
                f'{execution_context.as_posix()}/output.feature',
            ]
        )
        arguments.file = ' '.join(arguments.file)

        assert run(arguments, local_mock) == 0

        distributed_mock.assert_not_called()
        local_mock.assert_not_called()

        capture = capsys.readouterr()
        assert capture.err == ''
        assert capture.out == ''

        output_file = execution_context / 'output.feature'
        assert (
            output_file.read_text()
            == """Feature: a feature
    Scenario: first
        Given a variable with value "{{ hello }}"

    Scenario: second
        Given a variable with value "{{ s1foobar }}"
        Then run a bloody test
            \"\"\"
            with step text
            that spans
            more than
            one line
            \"\"\"

    # Scenario: third
    #     {% scenario "inactive-second", feature="./second.feature", prefix="s1" %}

    Scenario: third
        Given a variable with value "{{ s1value }}"
        Then run a bloody test, with table
          | hello | world |
          | foo   | bar   |
          | bar   |       |
          |       | foo   |
"""
        )
        assert (
            feature_file.read_text()
            == """Feature: a feature
    Scenario: first
        Given a variable with value "{{ hello }}"

    Scenario: second
        {% scenario "second", feature="../second.feature", prefix="s1" %}

    # Scenario: third
    #     {% scenario "inactive-second", feature="./second.feature", prefix="s1" %}

    Scenario: third
        {% scenario "third", feature="../second.feature", prefix="s1" %}
"""
        )

        feature_file.write_text("""Feature: a feature
    Scenario: first
        Given a variable with value "{{ hello }}"

    Scenario: second
        {% scenario "second", feature="../second.feature", foo="bar" %}
""")
        feature_file_2 = execution_context / 'second.feature'
        feature_file_2.write_text("""Feature: a second feature
    Background: common
        Given a common step

    Scenario: second
        Given a variable with value "{{ {$ bar $}foobar }}"
        Then run a bloody test
            \"\"\"
            with step text
            that spans
            more than
            one line
            \"\"\"
""")

        output_file.unlink(missing_ok=True)

        arguments = parser.parse_args(
            [
                'run',
                '-e',
                f'{execution_context.as_posix()}/configuration.yaml',
                '--yes',
                f'{execution_context.as_posix()}/features/test.feature',
                '--dump',
                f'{execution_context.as_posix()}/output.feature',
            ]
        )
        arguments.file = ' '.join(arguments.file)

        with pytest.raises(ValueError) as ve:  # noqa: PT011
            run(arguments, local_mock)

        assert (
            str(ve.value)
            == """the following variables has been declared in scenario tag but not used in ../second.feature#second:
  foo

the following variables was used in ../second.feature#second but was not declared in scenario tag:
  bar
"""
        )

        distributed_mock.assert_not_called()
        local_mock.assert_not_called()

        capture = capsys.readouterr()
        assert capture.err == ''
        assert capture.out == ''

        feature_file.write_text("""Feature: a feature
    Scenario: first
        Given a variable with value "{{ hello }}"

    Scenario: second
        {% scenario "second", feature="../second.feature", foo="bar" %}
""")
        feature_file_2 = execution_context / 'second.feature'
        feature_file_2.write_text("""Feature: a second feature
    Background: common
        Given a common step

    Scenario: second
        Given a variable with value "{{ {$ foo $}foobar }}"
        {% scenario "fourth", feature="./features/fourth.feature", foo="{$ foo $}", bar="foo", condition=True %}

        Then run a bloody test
            \"\"\"
            with step text
            that spans
            more than
            one line
            \"\"\"
""")

        feature_file_3.write_text("""Feature: a fourth feature
    Background: common
        Given a common step

    Scenario: fourth
        Then could it be "{$ foo $}" and "{$ bar $}"

        # <!-- this step is conditional -->
        {%- if {$ condition $} %}
        Then alert me!
        {%- endif %}
""")

        output_file.unlink(missing_ok=True)

        arguments = parser.parse_args(
            [
                'run',
                '-e',
                f'{execution_context.as_posix()}/configuration.yaml',
                '--yes',
                f'{execution_context.as_posix()}/features/test.feature',
                '--dump',
                f'{execution_context.as_posix()}/output.feature',
            ]
        )
        arguments.file = ' '.join(arguments.file)

        assert run(arguments, local_mock) == 0

        distributed_mock.assert_not_called()
        local_mock.assert_not_called()

        capture = capsys.readouterr()
        assert capture.err == ''
        assert capture.out == ''
        assert (
            output_file.read_text()
            == """Feature: a feature
    Scenario: first
        Given a variable with value "{{ hello }}"

    Scenario: second
        Given a variable with value "{{ barfoobar }}"
        Then could it be "bar" and "foo"

        # <!-- this step is conditional -->
        Then alert me!

        Then run a bloody test
            \"\"\"
            with step text
            that spans
            more than
            one line
            \"\"\"
"""
        )

        feature_file_2.write_text("""Feature: a second feature
    Background: common
        Given a common step

    Scenario: second
        Given a variable with value "{{ {$ foo $}foobar }}"
        {% scenario "fourth", feature="./features/fourth.feature", foo="{$ foo $}", bar="foo", condition=False %}

        Then run a bloody test
            \"\"\"
            with step text
            that spans
            more than
            one line
            \"\"\"
""")

        output_file.unlink(missing_ok=True)

        arguments = parser.parse_args(
            [
                'run',
                '-e',
                f'{execution_context.as_posix()}/configuration.yaml',
                '--yes',
                f'{execution_context.as_posix()}/features/test.feature',
                '--dump',
                f'{execution_context.as_posix()}/output.feature',
            ]
        )
        arguments.file = ' '.join(arguments.file)

        assert run(arguments, local_mock) == 0

        distributed_mock.assert_not_called()
        local_mock.assert_not_called()

        capture = capsys.readouterr()
        assert capture.err == ''
        assert capture.out == ''
        assert (
            output_file.read_text()
            == """Feature: a feature
    Scenario: first
        Given a variable with value "{{ hello }}"

    Scenario: second
        Given a variable with value "{{ barfoobar }}"
        Then could it be "bar" and "foo"

        # <!-- this step is conditional -->

        Then run a bloody test
            \"\"\"
            with step text
            that spans
            more than
            one line
            \"\"\"
"""
        )

        feature_file_3.write_text("""Feature: a fourth feature
    Background: common
        Given a common step

    Scenario: fourth
        Then could it be "{$ foo $}" and "{$ bar $}"
        \"\"\"
        hello world
        \"\"\"

        # <!-- this step is conditional -->
        {%- if {$ condition $} %}
        Then alert me!
        {%- endif %}

    Scenario: fifth
        Given a scenario after the included scenario
""")

        output_file.unlink(missing_ok=True)

        assert run(arguments, distributed_mock) == 0

        distributed_mock.assert_not_called()
        local_mock.assert_not_called()

        capture = capsys.readouterr()
        assert capture.err == ''
        assert capture.out == ''
        assert (
            output_file.read_text()
            == """Feature: a feature
    Scenario: first
        Given a variable with value "{{ hello }}"

    Scenario: second
        Given a variable with value "{{ barfoobar }}"
        Then could it be "bar" and "foo"
        \"\"\"
        hello world
        \"\"\"

        # <!-- this step is conditional -->

        Then run a bloody test
            \"\"\"
            with step text
            that spans
            more than
            one line
            \"\"\"
"""
        )
    finally:
        rm_rf(test_context)


def test_if_condition_with_scenario_tag_ext(caplog: LogCaptureFixture) -> None:
    environment = Environment(autoescape=False, extensions=[ScenarioTag])

    template = environment.from_string('{% if False %}hello {{ name }}!{% endif %}')
    with caplog.at_level(logging.DEBUG):
        assert template.render() == ''

    template = environment.from_string('{% if True %}hello {{ name }}!{% endif %}')
    assert template.render() == 'hello {{ name }}!'

    template = environment.from_string("""foobar

{%- if True %}
hello {{ name }}!
{%- endif %}
world""")
    assert template.render() == 'foobar\nhello {{ name }}!\nworld'

    template = environment.from_string("""Feature: a fourth feature
    Background: common
        Given a common step

    Scenario: fourth
        Then could it be "{$ foo $}" and "{$ bar $}"

        # <!-- this step is conditional -->
        {%- if True %}
        Then show me the "{{ money }}"
        {%- endif %}""")

    assert (
        template.render()
        == """Feature: a fourth feature
    Background: common
        Given a common step

    Scenario: fourth
        Then could it be "{$ foo $}" and "{$ bar $}"

        # <!-- this step is conditional -->
        Then show me the "{{ money }}"
        """.rstrip()
    )

    template = environment.from_string("""Feature: a fourth feature
    Background: common
        Given a common step

    Scenario: fourth
        Then could it be "{$ foo $}" and "{$ bar $}"

        # <!-- this step is conditional -->
        {%- if False %}
        Then show me the "{{ money }}"
        {%- endif %}""")

    assert (
        template.render()
        == """Feature: a fourth feature
    Background: common
        Given a common step

    Scenario: fourth
        Then could it be "{$ foo $}" and "{$ bar $}"

        # <!-- this step is conditional -->"""
    )



"""Tests for grizzly_cli.argparse.bashcompletion."""

from __future__ import annotations

import argparse
import inspect
from os.path import sep
from pathlib import Path
from typing import TYPE_CHECKING

import pytest
from grizzly_cli.__main__ import _create_parser
from grizzly_cli.argparse import ArgumentParser
from grizzly_cli.argparse.bashcompletion import BashCompleteAction, BashCompletionAction, hook

from test_cli.helpers import cwd, rm_rf

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Generator

    from _pytest.capture import CaptureFixture, CaptureResult
    from _pytest.tmpdir import TempPathFactory
    from pytest_mock import MockerFixture


@pytest.fixture
def test_parser() -> ArgumentParser:
    parser = ArgumentParser(prog='test-prog')
    parser.add_argument('--verbose', action='store_true')
    parser.add_argument('--file', action='append', type=str, required=True)
    parser.add_argument('test', nargs=1, type=str)
    parser.add_argument('--value', type=int)

    subparsers = parser.add_subparsers(dest='subparser')
    subparsers.add_parser('aparser')
    subparsers.add_parser('bparser')

    group = parser.add_mutually_exclusive_group()
    group.add_argument('--test', action='store_true')
    group.add_argument('--foo', action='store_true')
    group.add_argument('--bar', action='store_true')

    return parser


@pytest.fixture
def test_file_structure(tmp_path_factory: TempPathFactory) -> Generator[str, None, None]:
    test_context = tmp_path_factory.mktemp('test_context')
    file = test_context / 'test.txt'
    file.write_text('test.txt file')

    file = test_context / 'test.json'
    file.write_text('{"value": "test.json file"}')

    file = test_context / 'test.xml'
    file.write_text('<value>test.xml file</value>')

    file = test_context / 'test.yaml'
    file.write_text('test:')

    file = test_context / 'test space.yaml'
    file.write_text('test:')

    file = test_context / 'test.feature'
    file.write_text('Feature:')

    test_dir = test_context / 'test-dir'
    test_dir.mkdir()
    file = test_dir / 'test.yaml'
    file.write_text('test:')

    file = test_dir / 'test.feature'
    file.write_text('Feature:')

    hidden_dir = test_context / '.hidden'
    hidden_dir.mkdir()
    file = hidden_dir / 'hidden.txt'
    file.write_text('hidden.txt file')

    try:
        with cwd(test_context):
            yield str(test_context)
    finally:
        rm_rf(test_context)


class TestBashCompletionAction:
    def test___init__(self) -> None:
        action = BashCompletionAction(['--bash-completion'])

        assert isinstance(action, argparse.Action)
        assert action.dest == argparse.SUPPRESS
        assert action.default == argparse.SUPPRESS
        assert action.nargs == 0
        assert action.help == argparse.SUPPRESS

    def test___call__(self, capsys: CaptureFixture) -> None:
        parser = argparse.ArgumentParser(prog='test-prog')
        action = BashCompletionAction(['--bash-completion'])

        with pytest.raises(SystemExit) as e:
            action(parser, parser.parse_args([]), None)
        assert e.type is SystemExit
        assert e.value.code == 0

        bash_script_path = Path(inspect.getfile(action.__class__)).parent / 'bashcompletion.bash'

        with bash_script_path.open(encoding='utf-8') as fd:
            bash_script = fd.read().replace('bashcompletion_template', parser.prog) + '\n'

        capture = capsys.readouterr()
        assert capture.err == ''
        assert capture.out == bash_script


class TestBashCompleteAction:
    def test___init__(self) -> None:
        action = BashCompleteAction(['--bash-complete'])

        assert isinstance(action, argparse.Action)
        assert action.dest == argparse.SUPPRESS
        assert action.default == argparse.SUPPRESS
        assert action.nargs is None
        assert action.help == argparse.SUPPRESS

    def test_get_suggestions(self) -> None:
        parser = ArgumentParser(prog='test-prog')
        action = BashCompleteAction(['--bash-complete'])

        suggestions = action.get_suggestions(parser)
        assert list(suggestions.keys()) == ['-h', '--help']
        assert isinstance(suggestions.get('--help', None), argparse.Action)

        parser.add_argument('--verbose', action='store_true')
        parser.add_argument('--file', type=str, required=True)
        parser.add_argument('test', nargs=1, type=str)

        subparsers = parser.add_subparsers(dest='subparser')

        subparsers.add_parser('a')
        subparsers.add_parser('b')

        suggestions = action.get_suggestions(parser)
        assert list(suggestions.keys()) == ['-h', '--help', '--verbose', '--file', 'test', 'a', 'b']
        assert isinstance(suggestions.get('--verbose', None), argparse._StoreTrueAction)
        assert isinstance(suggestions.get('--file', None), argparse._StoreAction)
        assert isinstance(suggestions.get('test', None), argparse._StoreAction)
        assert isinstance(suggestions.get('a', None), argparse._SubParsersAction)
        assert isinstance(suggestions.get('b', None), argparse._SubParsersAction)

    def test_get_exclusive_suggestions(self, test_parser: ArgumentParser) -> None:
        parser = ArgumentParser(prog='test-prog')
        action = BashCompleteAction(['--bash-complete'])

        assert action.get_exclusive_suggestions(parser) == {}

        assert action.get_exclusive_suggestions(test_parser) == {
            '--test': ['--foo', '--bar'],
            '--foo': ['--test', '--bar'],
            '--bar': ['--test', '--foo'],
        }

    def test_get_provided_options(self) -> None:
        action = BashCompleteAction(['--bash-complete'])

        assert action.get_provided_options('test-prog', None) == []
        assert action.get_provided_options('test-prog', []) == []
        assert action.get_provided_options('test-prog', '') == []
        assert action.get_provided_options('test-prog', 'test-prog') == []
        assert action.get_provided_options('test-prog', 'test-prog --foo hello --bar') == ['--foo', 'hello', '--bar']
        assert action.get_provided_options('test-prog', ['test-prog', '--foo', 'hello', '--bar']) == ['--foo', 'hello', '--bar']

    def test_remove_completed(self, test_parser: ArgumentParser) -> None:
        action = BashCompleteAction(['--bash-complete'])

        suggestions = action.get_suggestions(test_parser)
        all_suggestions = suggestions.copy()
        all_options_sorted = sorted(all_suggestions.keys())
        exclusive_suggestions = action.get_exclusive_suggestions(test_parser)

        assert action.remove_completed([], suggestions, exclusive_suggestions) == []
        assert sorted(suggestions.keys()) == all_options_sorted
        assert action.remove_completed(['--verbose'], suggestions, exclusive_suggestions) == ['--verbose']
        assert sorted(suggestions.keys()) == all_options_sorted

        assert action.remove_completed(['--verbose', '--file'], suggestions, exclusive_suggestions) == ['--file']
        assert sorted([*suggestions.keys(), '--verbose']) == all_options_sorted

        assert action.remove_completed(['--verbose', '--file', 'test.txt'], suggestions, exclusive_suggestions) == []
        assert sorted([*suggestions.keys(), '--verbose']) == all_options_sorted

        assert action.remove_completed(['--verbose', '--file', 'test.txt', 'a'], suggestions, exclusive_suggestions) == ['a']
        assert sorted([*suggestions.keys(), '--verbose']) == all_options_sorted

        # if subparsers are completed, then we move to another parser, with its own arguments

        assert action.remove_completed(['--verbose', '--file', 'test.txt', '--value'], suggestions, exclusive_suggestions) == ['--value']
        assert sorted([*suggestions.keys(), '--verbose']) == all_options_sorted

        assert action.remove_completed(['--verbose', '--file', 'test.txt', '--value', '8'], suggestions, exclusive_suggestions) == ['--value', '8']
        assert sorted([*suggestions.keys(), '--verbose']) == all_options_sorted

        # only one of --foo, --bar, --test is valid (mutually exclusive), so all should be removed from suggestions if one of them is specified
        assert action.remove_completed(['--verbose', '--file', 'test.txt', '--value', '8', '--foo'], suggestions, exclusive_suggestions) == []
        assert sorted([*suggestions.keys(), '--verbose', '--value', '--foo', '--bar', '--test']) == all_options_sorted

    def test_filter_suggestions(self, test_parser: ArgumentParser) -> None:
        action = BashCompleteAction(['--bash-complete'])

        suggestions = action.get_suggestions(test_parser)
        all_suggestions = suggestions.copy()

        assert action.filter_suggestions([], suggestions) == all_suggestions
        assert sorted(action.filter_suggestions(['--'], suggestions).keys()) == sorted(['--help', '--test', '--foo', '--bar', '--value', '--verbose', '--file'])
        assert sorted(action.filter_suggestions(['--v'], suggestions).keys()) == sorted(['--verbose', '--value'])
        assert sorted(action.filter_suggestions(['--f'], suggestions).keys()) == sorted(['--file', '--foo'])

    @pytest.mark.parametrize(
        ('command', 'expected'),
        [
            ('grizzly-cli ', '-h\n--help\n--version\ninit\nkeyvault\nlocal\ndist\nauth'),
            ('grizzly-cli -', '-h\n--help\n--version'),
            ('grizzly-cli --', '--help\n--version'),
            ('grizzly-cli lo', 'local'),
            ('grizzly-cli -h', ''),
        ],
    )
    def test___call__(self, command: str, expected: str, capsys: CaptureFixture) -> None:
        parser = _create_parser()

        with pytest.raises(SystemExit):
            parser.parse_args([f'--bash-complete={command}'])
        capture = capsys.readouterr()
        assert sorted(capture.out.split('\n')) == sorted(f'{expected}\n'.split('\n'))

    @pytest.mark.parametrize(
        ('command', 'expected'),
        [
            (
                'grizzly-cli local run ',
                (
                    '-h\n--help\n--verbose\n-T\n--testdata-variable\n-y\n--yes\n-e\n--environment-file\n--csv-prefix\n--csv-interval\n'
                    '--csv-flush-interval\n-l\n--log-file\n--log-dir\n--dump\n--dry-run\ntest.feature\ntest-dir'
                ),
            ),
            (
                'grizzly-cli local run -',
                (
                    '-h\n--help\n--verbose\n-T\n--testdata-variable\n-y\n--yes\n-e\n--environment-file\n--csv-prefix\n--csv-interval\n--csv-flush-interval\n-l\n--log-file\n'
                    '--log-dir\n--dump\n--dry-run'
                ),
            ),
            (
                'grizzly-cli local run --',
                '--help\n--verbose\n--testdata-variable\n--yes\n--environment-file\n--csv-prefix\n--csv-interval\n--csv-flush-interval\n--log-file\n--log-dir\n--dump\n--dry-run',
            ),
            (
                'grizzly-cli local run --yes',
                (
                    '-h\n--help\n--verbose\n-T\n--testdata-variable\n-e\n--environment-file\n--csv-prefix\n--csv-interval\n--csv-flush-interval\n-l\n--log-file\n--log-dir\n'
                    '--dump\n--dry-run'
                ),
            ),
            ('grizzly-cli local run --help --yes', ''),
            ('grizzly-cli local run --yes -T', ''),
            (
                'grizzly-cli local run --yes -T key=value',
                (
                    '-h\n--help\n--verbose\n-T\n--testdata-variable\n-e\n--environment-file\n--csv-prefix\n--csv-interval\n'
                    '--csv-flush-interval\n-l\n--log-file\n--log-dir\n--dump\n--dry-run\ntest.feature\ntest-dir'
                ),
            ),
            ('grizzly-cli local run --yes -T key=value --env', '--environment-file'),
            ('grizzly-cli local run --yes -T key=value --environment-file', 'test.yaml\ntest\\ space.yaml\ntest-dir'),
            ('grizzly-cli local run --yes -T key=value --environment-file test', 'test.yaml\ntest\\ space.yaml\ntest-dir'),
            ('grizzly-cli local run --yes -T key=value --environment-file test-', 'test-dir'),
            (
                'grizzly-cli local run --yes -T key=value --environment-file test-dir',
                '-h\n--help\n--verbose\n-T\n--testdata-variable\n--csv-prefix\n--csv-interval\n--csv-flush-interval\n-l\n--log-file\n--log-dir\n--dump\n--dry-run',
            ),
            ('grizzly-cli local run --yes -T key=value --environment-file test.', 'test.yaml'),
            (
                'grizzly-cli local run --yes -T key=value --environment-file test.yaml',
                '-h\n--help\n--verbose\n-T\n--testdata-variable\n--csv-prefix\n--csv-interval\n--csv-flush-interval\n-l\n--log-file\n--log-dir\n--dump\n--dry-run',
            ),
            ('grizzly-cli local run --yes -T key=value --environment-file test.yaml --test', '--testdata-variable'),
            ('grizzly-cli local run --yes -T key=value --environment-file test.yaml --testdata-variable', ''),
            (
                'grizzly-cli local run --yes -T key=value --environment-file test.yaml --testdata-variable key=value',
                (
                    '-h\n--help\n--verbose\n-T\n--testdata-variable\n--csv-prefix\n--csv-interval\n--csv-flush-interval\n-l\n--log-file\n--log-dir\n--dump\n--dry-run\n'
                    'test.feature\ntest-dir'
                ),
            ),
            ('grizzly-cli local run --yes -T key=value --environment-file test.yaml --testdata-variable key=value test', 'test.feature\ntest-dir'),
            ('grizzly-cli local run --yes -T key=value --environment-file test.yaml --testdata-variable key=value test-dir', 'test-dir'),
            (f'grizzly-cli local run --yes -T key=value --environment-file test.yaml --testdata-variable key=value test-dir{sep}', f'test-dir{sep}test.feature'),
            (
                f'grizzly-cli local run --yes -T key=value --environment-file test.yaml --testdata-variable key=value test-dir{sep}tes',
                f'test-dir{sep}test.feature',
            ),
            (
                f'grizzly-cli local run --yes -T key=value --environment-file test.yaml --testdata-variable key=value test-dir{sep}test.feature',
                '-h\n--help\n--verbose\n-T\n--testdata-variable\n--csv-prefix\n--csv-interval\n--csv-flush-interval\n-l\n--log-file\n--log-dir\n--dump\n--dry-run',
            ),
            ('grizzly-cli local run --yes -T key=value --environment-file test.yaml --testdata-variable key=value test.fe', 'test.feature'),
            ('grizzly-cli local run --yes -T key=value --environment-file test.yaml --testdata-variable key=value --help', ''),
            ('grizzly-cli local run --yes -T key=value --environment-file test.yaml --testdata-variable key=value --help d', ''),
        ],
    )
    def test___call___local_run(self, command: str, expected: str, capsys: CaptureFixture, test_file_structure: str) -> None:  # noqa: ARG002
        capture: CaptureResult | None = None

        try:
            parser = _create_parser()
            hook(parser)
            _subparsers = getattr(parser, '_subparsers', None)
            assert _subparsers is not None
            subparser: argparse.ArgumentParser | None = None
            for subparsers in _subparsers._group_actions:
                for name, possible_subparser in subparsers.choices.items():
                    if name == 'local':
                        subparser = possible_subparser
                        break

            assert subparser is not None
            assert subparser.prog == 'grizzly-cli local'

            _subparsers = getattr(subparser, '_subparsers', None)
            assert _subparsers is not None
            subparser = None
            for subparsers in _subparsers._group_actions:
                for name, possible_subparser in subparsers.choices.items():
                    if name == 'run':
                        subparser = possible_subparser
                        break

            assert subparser is not None
            assert subparser.prog == 'grizzly-cli local run'

            with pytest.raises(SystemExit):
                subparser.parse_args([f'--bash-complete={command}'])
            capture = capsys.readouterr()
            actual = capture.out.rstrip()
            assert set(expected.split('\n')).issubset(actual.split('\n'))
        except:
            print(f'input={command}')
            print(f'expected={expected}')
            if capture is not None:
                print(f'actual={capture.out}')
            raise

    @pytest.mark.parametrize(
        ('command', 'expected'),
        [
            (
                'grizzly-cli dist run ',
                (
                    '-h\n--help\n--verbose\n-T\n--testdata-variable\n-y\n--yes\n-e\n--environment-file\n'
                    '--csv-prefix\n--csv-interval\n--csv-flush-interval\n-l\n--log-file\n--log-dir\n--dump\n--dry-run\ntest.feature\ntest-dir'
                ),
            ),
            (
                'grizzly-cli dist run -',
                (
                    '-h\n--help\n--verbose\n-T\n--testdata-variable\n-y\n--yes\n-e\n--environment-file\n--csv-prefix\n--csv-interval\n--csv-flush-interval\n-l\n'
                    '--log-file\n--log-dir\n--dump\n--dry-run'
                ),
            ),
            (
                'grizzly-cli dist run --',
                '--help\n--verbose\n--testdata-variable\n--yes\n--environment-file\n--csv-prefix\n--csv-interval\n--csv-flush-interval\n--log-file\n--log-dir\n--dump\n--dry-run',
            ),
            (
                'grizzly-cli dist run --yes',
                (
                    '-h\n--help\n--verbose\n-T\n--testdata-variable\n-e\n--environment-file\n--csv-prefix\n--csv-interval\n--csv-flush-interval\n'
                    '-l\n--log-file\n--log-dir\n--dump\n--dry-run'
                ),
            ),
            ('grizzly-cli dist run --help --yes', ''),
            ('grizzly-cli dist run --yes -T', ''),
            (
                'grizzly-cli dist run --yes -T key=value',
                (
                    '-h\n--help\n--verbose\n-T\n--testdata-variable\n-e\n--environment-file\n--csv-prefix\n--csv-interval\n'
                    '--csv-flush-interval\n-l\n--log-file\n--log-dir\n--dump\n--dry-run\ntest.feature\ntest-dir'
                ),
            ),
            ('grizzly-cli dist run --yes -T key=value --env', '--environment-file'),
            ('grizzly-cli dist run --yes -T key=value --environment-file', 'test.yaml\ntest\\ space.yaml\ntest-dir'),
            ('grizzly-cli dist run --yes -T key=value --environment-file test', 'test.yaml\ntest\\ space.yaml\ntest-dir'),
            ('grizzly-cli dist run --yes -T key=value --environment-file test-', 'test-dir'),
            (
                'grizzly-cli dist run --yes -T key=value --environment-file test-dir',
                '-h\n--help\n--verbose\n-T\n--testdata-variable\n--csv-prefix\n--csv-interval\n--csv-flush-interval\n-l\n--log-file\n--log-dir\n--dump\n--dry-run',
            ),
            ('grizzly-cli dist run --yes -T key=value --environment-file test.', 'test.yaml'),
            (
                'grizzly-cli dist run --yes -T key=value --environment-file test.yaml',
                '-h\n--help\n--verbose\n-T\n--testdata-variable\n--csv-prefix\n--csv-interval\n--csv-flush-interval\n-l\n--log-file\n--log-dir\n--dump\n--dry-run',
            ),
            ('grizzly-cli dist run --yes -T key=value --environment-file test.yaml --test', '--testdata-variable'),
            ('grizzly-cli dist run --yes -T key=value --environment-file test.yaml --testdata-variable', ''),
            (
                'grizzly-cli dist run --yes -T key=value --environment-file test.yaml --testdata-variable key=value',
                (
                    '-h\n--help\n--verbose\n-T\n--testdata-variable\n--csv-prefix\n--csv-interval\n--csv-flush-interval\ntest.feature\ntest-dir\n-l\n--log-file\n'
                    '--log-dir\n--dump\n--dry-run'
                ),
            ),
            ('grizzly-cli dist run --yes -T key=value --environment-file test.yaml --testdata-variable key=value test', 'test.feature\ntest-dir'),
            ('grizzly-cli dist run --yes -T key=value --environment-file test.yaml --testdata-variable key=value test.fe', 'test.feature'),
            ('grizzly-cli dist run --yes -T key=value --environment-file test.yaml --testdata-variable key=value --help', ''),
            ('grizzly-cli dist run --yes -T key=value --environment-file test.yaml --testdata-variable key=value --help d', ''),
        ],
    )
    def test___call___dist_run(self, command: str, expected: str, capsys: CaptureFixture, test_file_structure: str) -> None:  # noqa: ARG002
        capture: CaptureResult | None = None

        try:
            parser = _create_parser()
            hook(parser)
            _subparsers = getattr(parser, '_subparsers', None)
            assert _subparsers is not None
            subparser: argparse.ArgumentParser | None = None
            for subparsers in _subparsers._group_actions:
                for name, possible_subparser in subparsers.choices.items():
                    if name == 'dist':
                        subparser = possible_subparser
                        break

            assert subparser is not None
            assert subparser.prog == 'grizzly-cli dist'

            _subparsers = getattr(subparser, '_subparsers', None)
            assert _subparsers is not None
            subparser = None
            for subparsers in _subparsers._group_actions:
                for name, possible_subparser in subparsers.choices.items():
                    if name == 'run':
                        subparser = possible_subparser
                        break

            assert subparser is not None
            assert subparser.prog == 'grizzly-cli dist run'

            with pytest.raises(SystemExit):
                subparser.parse_args([f'--bash-complete={command}'])
            capture = capsys.readouterr()
            actual = capture.out.rstrip()
            assert set(expected.split('\n')).issubset(actual.split('\n'))
        except:
            print(f'input={command}')
            print(f'expected={expected}')
            if capture is not None:
                print(f'actual={capture.out}')
            raise

    @pytest.mark.parametrize(
        ('command', 'expected'),
        [
            (
                'grizzly-cli dist',
                (
                    '-h\n--help\n--workers\n--id\n--limit-nofile\n--health-retries\n--health-timeout\n--health-interval\n--registry\n'
                    '--tty\n--wait-for-worker\n--project-name\n--force-build\n--build\n--validate-config\nbuild\nclean\nrun'
                ),
            ),
            (
                'grizzly-cli dist -',
                (
                    '-h\n--help\n--workers\n--id\n--limit-nofile\n--health-retries\n--health-timeout\n--health-interval\n--registry\n'
                    '--tty\n--wait-for-worker\n--project-name\n--force-build\n--build\n--validate-config'
                ),
            ),
            (
                'grizzly-cli dist --',
                (
                    '--help\n--workers\n--id\n--limit-nofile\n--health-retries\n--health-timeout\n--health-interval\n--registry\n'
                    '--tty\n--wait-for-worker\n--project-name\n--force-build\n--build\n--validate-config'
                ),
            ),
            (
                'grizzly-cli dist --workers',
                '',
            ),
            (
                'grizzly-cli dist --workers asdf',
                '',
            ),
            (
                'grizzly-cli dist --workers 8',
                (
                    '-h\n--help\n--id\n--limit-nofile\n--health-retries\n--health-timeout\n--health-interval\n--registry\n--tty\n'
                    '--wait-for-worker\n--project-name\n--force-build\n--build\n--validate-config\nbuild\nclean\nrun'
                ),
            ),
            (
                'grizzly-cli dist --workers 8 --force-build',
                (
                    '-h\n--help\n--id\n--limit-nofile\n--health-retries\n--health-timeout\n--health-interval\n--registry\n--tty\n'
                    '--wait-for-worker\n--project-name\nbuild\nclean\nrun'
                ),
            ),
        ],
    )
    def test___call__dist(self, command: str, expected: str, capsys: CaptureFixture, test_file_structure: str) -> None:  # noqa: ARG002
        capture: CaptureResult | None = None

        try:
            parser = _create_parser()
            hook(parser)
            _subparsers = getattr(parser, '_subparsers', None)
            assert _subparsers is not None
            subparser: argparse.ArgumentParser | None = None
            for subparsers in _subparsers._group_actions:
                for name, possible_subparser in subparsers.choices.items():
                    if name == 'dist':
                        subparser = possible_subparser
                        break

            assert subparser is not None
            assert subparser.prog == 'grizzly-cli dist'

            with pytest.raises(SystemExit):
                subparser.parse_args([f'--bash-complete={command}'])
            capture = capsys.readouterr()
            assert sorted(capture.out.split('\n')) == sorted(f'{expected}\n'.split('\n'))
        except:
            print(f'input={command}')
            print(f'expected={expected}')
            if capture is not None:
                print(f'actual={capture.out}')
            raise

    @pytest.mark.parametrize(
        ('command', 'expected'),
        [
            (
                'grizzly-cli dist build',
                '-h\n--help\n--no-cache\n--registry\n--no-progress\n--verbose',
            ),
            (
                'grizzly-cli dist build --',
                '--help\n--no-cache\n--registry\n--no-progress\n--verbose',
            ),
            (
                'grizzly-cli dist build --help',
                '',
            ),
            (
                'grizzly-cli dist build --no-cache',
                '-h\n--help\n--registry\n--no-progress\n--verbose',
            ),
            (
                'grizzly-cli dist build --no-cache --registry',
                '',
            ),
            (
                'grizzly-cli dist build --no-cache --registry asdf',
                '-h\n--help\n--no-progress\n--verbose',
            ),
        ],
    )
    def test___call__dist_build(self, command: str, expected: str, capsys: CaptureFixture, test_file_structure: str) -> None:  # noqa: ARG002
        capture: CaptureResult | None = None

        try:
            parser = _create_parser()
            hook(parser)
            _subparsers = getattr(parser, '_subparsers', None)
            assert _subparsers is not None
            subparser: argparse.ArgumentParser | None = None
            for subparsers in _subparsers._group_actions:
                for name, possible_subparser in subparsers.choices.items():
                    if name == 'dist':
                        subparser = possible_subparser
                        break

            assert subparser is not None
            assert subparser.prog == 'grizzly-cli dist'

            _subparsers = getattr(subparser, '_subparsers', None)
            assert _subparsers is not None
            subparser = None
            for subparsers in _subparsers._group_actions:
                for name, possible_subparser in subparsers.choices.items():
                    if name == 'build':
                        subparser = possible_subparser
                        break

            assert subparser is not None
            assert subparser.prog == 'grizzly-cli dist build'

            with pytest.raises(SystemExit):
                subparser.parse_args([f'--bash-complete={command}'])
            capture = capsys.readouterr()
            assert sorted(capture.out.split('\n')) == sorted(f'{expected}\n'.split('\n'))
        except:
            print(f'input={command}')
            print(f'expected={expected}')
            if capture is not None:
                print(f'actual={capture.out}')
            raise

    @pytest.mark.parametrize(
        ('command', 'expected'),
        [
            (
                'grizzly-cli dist clean',
                '-h\n--help\n--no-images\n--no-networks',
            ),
            (
                'grizzly-cli dist clean --no',
                '--no-images\n--no-networks',
            ),
        ],
    )
    def test___call__dist_clean(self, command: str, expected: str, capsys: CaptureFixture, test_file_structure: str) -> None:  # noqa: ARG002
        capture: CaptureResult | None = None

        try:
            parser = _create_parser()
            hook(parser)
            _subparsers = getattr(parser, '_subparsers', None)
            assert _subparsers is not None
            subparser: argparse.ArgumentParser | None = None
            for subparsers in _subparsers._group_actions:
                for name, possible_subparser in subparsers.choices.items():
                    if name == 'dist':
                        subparser = possible_subparser
                        break

            assert subparser is not None
            assert subparser.prog == 'grizzly-cli dist'

            _subparsers = getattr(subparser, '_subparsers', None)
            assert _subparsers is not None
            subparser = None
            for subparsers in _subparsers._group_actions:
                for name, possible_subparser in subparsers.choices.items():
                    if name == 'clean':
                        subparser = possible_subparser
                        break

            assert subparser is not None
            assert subparser.prog == 'grizzly-cli dist clean'

            with pytest.raises(SystemExit):
                subparser.parse_args([f'--bash-complete={command}'])
            capture = capsys.readouterr()
            assert sorted(capture.out.split('\n')) == sorted(f'{expected}\n'.split('\n'))
        except:
            print(f'input={command}')
            print(f'expected={expected}')
            if capture is not None:
                print(f'actual={capture.out}')
            raise


def test_hook(mocker: MockerFixture) -> None:
    parser = argparse.ArgumentParser(prog='test-prog')
    action = parser.add_argument('--test')

    assert len(parser._actions) == 2

    hook(parser)

    assert len(parser._actions) == 3

    option_strings = [option for action in parser._actions for option in action.option_strings]

    assert option_strings == ['-h', '--help', '--test', '--bash-complete']

    try:
        hook(parser)
    except argparse.ArgumentError as e:
        pytest.fail(str(e))

    subparsers = parser.add_subparsers(dest='test')
    subparser = subparsers.add_parser('test')

    assert len(subparser._actions) == 1

    hook(parser)

    assert len(subparser._actions) == 2

    option_strings = [option for action in subparser._actions for option in action.option_strings]

    assert option_strings == ['-h', '--help', '--bash-complete']

    mocker.patch.object(
        parser,
        'add_argument',
        side_effect=[
            argparse.ArgumentError(message='unrecognized arguments: --bash-completion', argument=action),
            RuntimeError('something else'),
        ],
    )

    with pytest.raises(argparse.ArgumentError) as ae:
        hook(parser)
    assert 'unrecognized arguments: --bash-completion' in ae.value.message

    with pytest.raises(RuntimeError) as re:
        hook(parser)
    assert 'something else' in str(re)



"""Tests for grizzly-cli --version."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from test_cli.helpers import get_current_version, rm_rf, run_command

if TYPE_CHECKING:
    from _pytest.tmpdir import TempPathFactory

CURRENT_VERSION, CURRENT_COMMON_VERSION = get_current_version()


@pytest.mark.parametrize(
    ('pip_module', 'grizzly_version', 'locust_version'),
    [
        ('grizzly-loadtester==1.0.0', '1.0.0', '2.2.1'),
        ('grizzly-loadtester[mq]==2.4.6', '2.4.6 ── extras: mq', '2.9.0'),
        ('git+https://git@github.com/biometria-se/grizzly.git@v1.4.1#egg=grizzly-loadtester', '0.0.0', '2.2.1'),
        ('git+https://git@github.com/biometria-se/grizzly.git@v2.4.6#egg=grizzly-loadtester', '0.0.0', '2.9.0'),
        ('git+https://git@github.com/biometria-se/grizzly.git@7285294b#egg=grizzly-loadtester', '2.4.7.dev7', '>=2.12.0,<2.13'),
        ('grizzly-loadtester[mq] @ git+https://git@github.com/biometria-se/grizzly.git@7285294b', '2.4.7.dev7 ── extras: mq', '>=2.12.0,<2.13'),
    ],
)
def test_e2e_version(pip_module: str, grizzly_version: str, locust_version: str, tmp_path_factory: TempPathFactory) -> None:
    test_context = tmp_path_factory.mktemp('test_context') / 'foobar'

    result: str | None = None

    try:
        # create project
        rc, output = run_command(
            ['grizzly-cli', 'init', 'foobar', '--yes'],
            cwd=test_context.parent,
        )

        try:
            assert rc == 0
        except AssertionError:
            print(''.join(output))
            raise

        requirements_file = test_context / 'requirements.txt'

        requirements_file.unlink()
        requirements_file.write_text(f'{pip_module}\n')

        rc, output = run_command(
            ['grizzly-cli', '--version', 'all'],
            cwd=test_context,
        )

        result = ''.join(output)

        expected = f"""grizzly-cli {CURRENT_VERSION}
├── grizzly-common {CURRENT_COMMON_VERSION}
└── grizzly {grizzly_version}
    └── locust {locust_version}
"""
        assert rc == 0
        assert expected in ''.join(output)
    except AssertionError:
        if result is not None:
            print('actual:')
            print(result)
            print('expected:')
            print(expected)
        raise
    finally:
        rm_rf(test_context)



"""Functionality for `grizzly-cli auth ...`."""

from __future__ import annotations

import sys
from os import environ
from pathlib import Path
from typing import TYPE_CHECKING

from pyotp import TOTP

from grizzly_cli import register_parser

if TYPE_CHECKING:  # pragma: no cover
    from argparse import Namespace as Arguments

    from grizzly_cli.argparse import ArgumentSubParser


@register_parser()
def create_parser(sub_parser: ArgumentSubParser) -> None:
    # grizzly-cli auth
    auth_parser = sub_parser.add_parser('auth', description=('grizzly stateless authenticator application'))

    auth_parser.add_argument(
        'input',
        nargs='?',
        type=str,
        default=None,
        const=None,
        help=('where to read OTP secret, nothing specified means environment variable OTP_SECRET, `-` means stdin and anything else is considered a file'),
    )

    if auth_parser.prog != 'grizzly-cli auth':  # pragma: no cover
        auth_parser.prog = 'grizzly-cli auth'


def auth(args: Arguments) -> int:
    secret: str | None = None

    if args.input is None:
        secret = environ.get('OTP_SECRET', None)
        if secret is None:
            message = 'environment variable OTP_SECRET is not set'
            raise ValueError(message)
    elif args.input == '-':
        try:
            secret = sys.stdin.read().strip()
        except:  # noqa: S110
            pass
        finally:
            if secret is None or len(secret.strip()) < 1:
                message = 'OTP secret could not be read from stdin'
                raise ValueError(message)
    else:
        input_file = Path(args.input)

        if not input_file.exists():
            message = f'file {input_file.as_posix()} does not exist'
            raise ValueError(message)

        secret = input_file.read_text().strip()

        if ' ' in secret or len(secret.split('\n')) > 1 or secret == '':
            message = f'file {input_file.as_posix()} does not seem to contain a single line with a valid OTP secret'
            raise ValueError(message)

    try:
        totp = TOTP(secret)

        print(totp.now())
    except Exception as e:
        message = f'unable to generate TOTP code: {e!s}'
        raise ValueError(message) from e

    return 0



"""Webserver with endpoints used during E2E testing."""

from __future__ import annotations

import csv
import logging
import socket
from contextlib import suppress
from pathlib import Path
from time import time
from typing import TYPE_CHECKING, Any, Literal

import gevent
import requests
from flask import Flask, jsonify, request
from flask import Response as FlaskResponse
from gevent.pywsgi import WSGIServer
from typing_extensions import Self

if TYPE_CHECKING:
    from types import TracebackType

logger = logging.getLogger('webserver')


app = Flask('webserver')


@app.route('/health')
def app_health() -> FlaskResponse:
    return jsonify({'status': 'ok'})


@app.route('/api/v1/resources/dogs')
def app_get_dog_fact() -> FlaskResponse:
    _ = int(request.args.get('number', ''))

    return jsonify(['woof woof wooof'])


@app.route('/facts')
def app_get_cat_fact() -> FlaskResponse:
    _ = int(request.args.get('limit', ''))

    return jsonify(['meow meow meow'])


@app.route('/books/<book>.json')
def app_get_book(book: str) -> FlaskResponse:
    books = Path.joinpath(Path.cwd(), 'features', 'requests', 'books', 'books.csv')
    with books.open('r') as fd:
        reader = csv.DictReader(fd)
        for row in reader:
            if row['book'] == book:
                return jsonify(
                    {
                        'number_of_pages': row['pages'],
                        'isbn_10': [row['isbn_10']] * 2,
                        'authors': [
                            {'key': '/author/' + row['author'].replace(' ', '_').strip() + '|' + row['isbn_10'].strip()},
                        ],
                    }
                )

    response = jsonify({'success': False})
    response.status_code = 500

    return response


@app.route('/author/<author_key>.json')
def app_get_author(author_key: str) -> FlaskResponse:
    name, _ = author_key.rsplit('|', 1)

    return jsonify(
        {
            'name': name.replace('_', ' '),
        }
    )


@app.errorhandler(404)
def catch_all(_: Any) -> FlaskResponse:
    return jsonify({}, status=200)


class Webserver:
    _web_server: WSGIServer

    def __init__(self, port: int = 0) -> None:
        self._web_server = WSGIServer(
            ('0.0.0.0', port),
            app,
            log=None,
        )
        logger.debug('created webserver on port %d', port)

    @property
    def port(self) -> int:
        port = self._web_server.server_port
        assert port is not None

        return port  # type: ignore[no-any-return]

    def wait_for_start(self, timeout: int = 10) -> None:
        start_time = time()
        while time() - start_time < timeout:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
                if sock.connect_ex(('127.0.0.1', self.port)) == 0:
                    return
                gevent.sleep(0.5)

        message = f'webserver did not start on port {self.port}'
        raise RuntimeError(message)

    def wait_for_health(self, timeout: int = 300) -> None:
        start_time = time()
        while time() - start_time < timeout:
            with suppress(Exception):
                response = requests.get(f'http://127.0.0.1:{self.port}/health', timeout=2.0)
                if response.status_code == 200:
                    return

            logger.error('webserver not healthy yet, retrying...')

            gevent.sleep(1.0)

        message = 'webserver did not start responding to requests in due time'
        raise RuntimeError(message)

    def start(self, logger_: logging.Logger | None = None) -> None:
        if logger_ is not None:
            global logger  # noqa: PLW0603
            logger = logger_

        gevent.spawn(lambda: self._web_server.serve_forever())
        self.wait_for_start()
        self.wait_for_health()
        logger.debug('started webserver on port %d', self.port)

    def __enter__(self) -> Self:
        self.start()

        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        traceback: TracebackType | None,
    ) -> Literal[True]:
        self._web_server.stop_accepting()
        self._web_server.stop()

        logger.debug('stopped webserver on port %d', self.port)

        return True



"""Tests for grizzly_cli.distributed.build."""

from __future__ import annotations

import sys
from argparse import Namespace
from contextlib import suppress
from inspect import getfile
from os import environ
from pathlib import Path
from socket import gaierror
from typing import TYPE_CHECKING

import pytest
from grizzly_cli.distributed.build import _create_build_command, build, getgid, getuid
from grizzly_cli.utils import RunCommandResult, rm_rf

from test_cli.helpers import cwd

if TYPE_CHECKING:
    from _pytest.capture import CaptureFixture
    from _pytest.tmpdir import TempPathFactory
    from pytest_mock import MockerFixture


@pytest.mark.skipif(sys.platform == 'win32', reason='only run test on non-windows platforms')
def test_getuid_getgid_win32(mocker: MockerFixture) -> None:
    mocker.patch('grizzly_cli.distributed.build.sys.platform', 'win32')

    assert getuid() == 1000
    assert getgid() == 1000

    mocker.patch('grizzly_cli.distributed.build.sys.platform', 'linux')

    assert getuid() >= 0
    assert getgid() >= 0


def test__create_build_command(mocker: MockerFixture) -> None:
    mocker.patch('grizzly_cli.distributed.build.getuid', return_value=1337)
    mocker.patch('grizzly_cli.distributed.build.getgid', return_value=2147483647)
    args = Namespace(container_system='test', local_install=False)

    mocker.patch('grizzly_cli.distributed.build.get_dependency_versions', return_value=(('1.1.1', None), '2.8.4'))

    assert _create_build_command(args, 'Containerfile.test', 'grizzly-cli:test', '/home/grizzly-cli/') == [
        'test',
        'image',
        'build',
        '--ssh',
        'default',
        '--build-arg',
        'GRIZZLY_EXTRA=base',
        '--build-arg',
        'GRIZZLY_INSTALL_TYPE=remote',
        '--build-arg',
        'GRIZZLY_UID=1337',
        '--build-arg',
        'GRIZZLY_GID=2147483647',
        '-f',
        'Containerfile.test',
        '-t',
        'grizzly-cli:test',
        '/home/grizzly-cli/',
    ]

    mocker.patch('grizzly_cli.distributed.build.get_dependency_versions', return_value=(('1.1.1', []), '2.8.4'))

    args.local_install = True

    assert _create_build_command(args, 'Containerfile.test', 'grizzly-cli:test', '/home/grizzly-cli/') == [
        'test',
        'image',
        'build',
        '--ssh',
        'default',
        '--build-arg',
        'GRIZZLY_EXTRA=base',
        '--build-arg',
        'GRIZZLY_INSTALL_TYPE=local',
        '--build-arg',
        'GRIZZLY_UID=1337',
        '--build-arg',
        'GRIZZLY_GID=2147483647',
        '-f',
        'Containerfile.test',
        '-t',
        'grizzly-cli:test',
        '/home/grizzly-cli/',
    ]

    mocker.patch('grizzly_cli.distributed.build.get_dependency_versions', return_value=(('1.1.1', ['dev', 'ci', 'mq']), '2.8.4'))

    assert _create_build_command(args, 'Containerfile.test', 'grizzly-cli:test', '/home/grizzly-cli/') == [
        'test',
        'image',
        'build',
        '--ssh',
        'default',
        '--build-arg',
        'GRIZZLY_EXTRA=mq',
        '--build-arg',
        'GRIZZLY_INSTALL_TYPE=local',
        '--build-arg',
        'GRIZZLY_UID=1337',
        '--build-arg',
        'GRIZZLY_GID=2147483647',
        '-f',
        'Containerfile.test',
        '-t',
        'grizzly-cli:test',
        '/home/grizzly-cli/',
    ]

    try:
        environ['IBM_MQ_LIB_HOST'] = 'https://localhost:8003'
        args.local_install = False

        assert _create_build_command(args, 'Containerfile.test', 'grizzly-cli:test', '/home/grizzly-cli/') == [
            'test',
            'image',
            'build',
            '--ssh',
            'default',
            '--build-arg',
            'GRIZZLY_EXTRA=mq',
            '--build-arg',
            'GRIZZLY_INSTALL_TYPE=remote',
            '--build-arg',
            'GRIZZLY_UID=1337',
            '--build-arg',
            'GRIZZLY_GID=2147483647',
            '--build-arg',
            'IBM_MQ_LIB_HOST=https://localhost:8003',
            '-f',
            'Containerfile.test',
            '-t',
            'grizzly-cli:test',
            '/home/grizzly-cli/',
        ]

        environ['IBM_MQ_LIB_HOST'] = 'http://host.docker.internal:8000'

        mocker.patch('grizzly_cli.distributed.build.gethostbyname', return_value='1.2.3.4')
        args.local_install = True

        assert _create_build_command(args, 'Containerfile.test', 'grizzly-cli:test', '/home/grizzly-cli/') == [
            'test',
            'image',
            'build',
            '--ssh',
            'default',
            '--build-arg',
            'GRIZZLY_EXTRA=mq',
            '--build-arg',
            'GRIZZLY_INSTALL_TYPE=local',
            '--build-arg',
            'GRIZZLY_UID=1337',
            '--build-arg',
            'GRIZZLY_GID=2147483647',
            '--build-arg',
            'IBM_MQ_LIB_HOST=http://host.docker.internal:8000',
            '--add-host',
            'host.docker.internal:1.2.3.4',
            '-f',
            'Containerfile.test',
            '-t',
            'grizzly-cli:test',
            '/home/grizzly-cli/',
        ]

        mocker.patch('grizzly_cli.distributed.build.gethostbyname', side_effect=[gaierror] * 2)

        assert _create_build_command(args, 'Containerfile.test', 'grizzly-cli:test', '/home/grizzly-cli/') == [
            'test',
            'image',
            'build',
            '--ssh',
            'default',
            '--build-arg',
            'GRIZZLY_EXTRA=mq',
            '--build-arg',
            'GRIZZLY_INSTALL_TYPE=local',
            '--build-arg',
            'GRIZZLY_UID=1337',
            '--build-arg',
            'GRIZZLY_GID=2147483647',
            '--build-arg',
            'IBM_MQ_LIB_HOST=http://host.docker.internal:8000',
            '--add-host',
            'host.docker.internal:host-gateway',
            '-f',
            'Containerfile.test',
            '-t',
            'grizzly-cli:test',
            '/home/grizzly-cli/',
        ]

        environ['IBM_MQ_LIB'] = 'mqm.tar.gz'

        args.local_install = False

        assert _create_build_command(args, 'Containerfile.test', 'grizzly-cli:test', '/home/grizzly-cli/') == [
            'test',
            'image',
            'build',
            '--ssh',
            'default',
            '--build-arg',
            'GRIZZLY_EXTRA=mq',
            '--build-arg',
            'GRIZZLY_INSTALL_TYPE=remote',
            '--build-arg',
            'GRIZZLY_UID=1337',
            '--build-arg',
            'GRIZZLY_GID=2147483647',
            '--build-arg',
            'IBM_MQ_LIB_HOST=http://host.docker.internal:8000',
            '--add-host',
            'host.docker.internal:host-gateway',
            '--build-arg',
            'IBM_MQ_LIB=mqm.tar.gz',
            '-f',
            'Containerfile.test',
            '-t',
            'grizzly-cli:test',
            '/home/grizzly-cli/',
        ]

    finally:
        with suppress(KeyError):
            del environ['IBM_MQ_LIB_HOST']

        with suppress(KeyError):
            del environ['IBM_MQ_LIB']


def test_build(capsys: CaptureFixture, mocker: MockerFixture, tmp_path_factory: TempPathFactory) -> None:  # noqa: PLR0915
    test_context = tmp_path_factory.mktemp('test_context')

    try:
        with cwd(test_context):
            mocker.patch('grizzly_cli.EXECUTION_CONTEXT', test_context.as_posix())
            mocker.patch('grizzly_cli.distributed.build.EXECUTION_CONTEXT', test_context.as_posix())
            mocker.patch('grizzly_cli.distributed.build.PROJECT_NAME', 'grizzly-scenarios')
            mocker.patch('grizzly_cli.distributed.build.getuser', return_value='test-user')
            mocker.patch('grizzly_cli.distributed.build.getuid', return_value=1337)
            mocker.patch('grizzly_cli.distributed.build.getgid', return_value=2147483647)
            run_command = mocker.patch(
                'grizzly_cli.distributed.build.run_command',
                side_effect=[
                    RunCommandResult(return_code=254),
                    RunCommandResult(return_code=133),
                    RunCommandResult(return_code=0),
                    RunCommandResult(return_code=1),
                    RunCommandResult(return_code=0),
                    RunCommandResult(return_code=0),
                    RunCommandResult(return_code=2),
                    RunCommandResult(return_code=0),
                    RunCommandResult(return_code=0),
                    RunCommandResult(return_code=0),
                ],
            )
            setattr(getattr(build, '__wrapped__'), '__value__', test_context.as_posix())  # noqa: B009, B010

            test_args = Namespace(container_system='test', force_build=False, project_name=None, local_install=False, no_progress=False, verbose=False)

            static_context = Path.joinpath(Path(getfile(_create_build_command)).parent, '..', 'static').resolve()

            assert build(test_args) == 254

            capture = capsys.readouterr()
            assert capture.err == ''
            assert capture.out == ''
            assert run_command.call_count == 1
            args, kwargs = run_command.call_args_list[-1]

            container_file_actual = args[0].pop(14)
            container_file_expected = Path.joinpath(static_context, 'Containerfile').as_posix()

            if sys.platform == 'win32':
                container_file_actual = container_file_actual.lower()
                container_file_expected = container_file_expected.lower()

            assert args[0] == [
                'test',
                'image',
                'build',
                '--ssh',
                'default',
                '--build-arg',
                'GRIZZLY_EXTRA=base',
                '--build-arg',
                'GRIZZLY_INSTALL_TYPE=remote',
                '--build-arg',
                'GRIZZLY_UID=1337',
                '--build-arg',
                'GRIZZLY_GID=2147483647',
                '-f',
                '-t',
                'grizzly-scenarios:test-user',
                test_context.as_posix(),
            ]

            assert container_file_actual == container_file_expected

            actual_env = kwargs.get('env', None)
            assert actual_env is not None
            assert actual_env.get('DOCKER_BUILDKIT', None) == environ.get('DOCKER_BUILDKIT', None)

            test_args = Namespace(container_system='docker', force_build=True, local_install=True, project_name='foobar', no_progress=False, verbose=False)

            mocker.patch('grizzly_cli.distributed.build.get_dependency_versions', return_value=(('1.1.1', ['mq', 'dev']), '2.8.4'))

            assert build(test_args) == 133
            assert run_command.call_count == 2
            args, kwargs = run_command.call_args_list[-1]

            container_file_actual = args[0].pop(14)
            container_file_expected = Path.joinpath(static_context, 'Containerfile').as_posix()

            if sys.platform == 'win32':
                container_file_actual = container_file_actual.lower()
                container_file_expected = container_file_expected.lower()

            assert args[0] == [
                'docker',
                'image',
                'build',
                '--ssh',
                'default',
                '--build-arg',
                'GRIZZLY_EXTRA=mq',
                '--build-arg',
                'GRIZZLY_INSTALL_TYPE=local',
                '--build-arg',
                'GRIZZLY_UID=1337',
                '--build-arg',
                'GRIZZLY_GID=2147483647',
                '-f',
                '-t',
                'foobar:test-user',
                test_context.as_posix(),
                '--no-cache',
            ]
            assert container_file_actual == container_file_expected

            capsys.readouterr()

            actual_env = kwargs.get('env', None)
            assert actual_env is not None
            assert actual_env.get('DOCKER_BUILDKIT', None) == '1'

            image_name = 'grizzly-scenarios:test-user'
            test_args = Namespace(
                container_system='docker',
                force_build=False,
                local_install=False,
                project_name=None,
                registry='ghcr.io/biometria-se/',
                no_progress=False,
                verbose=False,
            )

            assert build(test_args) == 1

            capture = capsys.readouterr()
            assert capture.err == ''
            assert capture.out == (f'\nbuilt image {image_name}\n\n!! failed to tag image {image_name} -> ghcr.io/biometria-se/{image_name}\n')

            assert run_command.call_count == 4

            args, kwargs = run_command.call_args_list[-1]
            assert args[0] == [
                'docker',
                'image',
                'tag',
                image_name,
                f'ghcr.io/biometria-se/{image_name}',
            ]

            actual_env = kwargs.get('env', None)
            assert actual_env.get('DOCKER_BUILDKIT', None) == '1'

            test_args = Namespace(
                container_system='docker',
                force_build=True,
                no_cache=True,
                build=True,
                registry='ghcr.io/biometria-se/',
                project_name='foobar',
                local_install=True,
                no_progress=False,
                verbose=False,
            )

            image_name = 'foobar:test-user'
            assert build(test_args) == 2

            capture = capsys.readouterr()
            assert capture.err == ''
            assert capture.out == (
                f'\nbuilt image {image_name}\ntagged image {image_name} -> ghcr.io/biometria-se/{image_name}\n\n!! failed to push image ghcr.io/biometria-se/{image_name}\n'
            )

            assert run_command.call_count == 7

            args, kwargs = run_command.call_args_list[-1]
            assert args[0] == [
                'docker',
                'image',
                'push',
                f'ghcr.io/biometria-se/{image_name}',
            ]

            actual_env = kwargs.get('env', None)
            assert actual_env.get('DOCKER_BUILDKIT', None) == '1'

            assert build(test_args) == 0

            capture = capsys.readouterr()
            assert capture.err == ''
            assert capture.out == (f'\nbuilt image {image_name}\ntagged image {image_name} -> ghcr.io/biometria-se/{image_name}\npushed image ghcr.io/biometria-se/{image_name}\n')
    finally:
        rm_rf(test_context)



"""Tests for grizzly_cli.local."""

from __future__ import annotations

from argparse import ArgumentParser, Namespace
from contextlib import suppress
from os import environ
from typing import TYPE_CHECKING

import pytest
from grizzly_cli.local import create_parser, local, local_run
from grizzly_cli.utils import RunCommandResult, rm_rf

if TYPE_CHECKING:
    from _pytest.tmpdir import TempPathFactory
    from pytest_mock import MockerFixture


def test_local(mocker: MockerFixture) -> None:
    run_mocked = mocker.patch('grizzly_cli.local.run', return_value=0)

    arguments = Namespace(subcommand='run')

    assert local(arguments) == 0
    assert run_mocked.call_count == 1
    args, _ = run_mocked.call_args_list[0]
    assert args[0] is arguments
    assert args[1] is local_run

    arguments = Namespace(subcommand='foo')
    with pytest.raises(ValueError, match='unknown subcommand foo'):
        local(arguments)


def test_local_run(mocker: MockerFixture, tmp_path_factory: TempPathFactory) -> None:
    run_command = mocker.patch('grizzly_cli.local.run_command', return_value=RunCommandResult(return_code=0))
    test_context = tmp_path_factory.mktemp('test_context')
    (test_context / 'test.feature').write_text('Feature:')

    parser = ArgumentParser()

    sub_parsers = parser.add_subparsers(dest='test')

    create_parser(sub_parsers)

    try:
        assert environ.get('GRIZZLY_TEST_VAR', None) is None

        arguments = parser.parse_args(
            [
                'local',
                'run',
                f'{test_context}/test.feature',
            ]
        )

        arguments.file = ' '.join(arguments.file)

        assert (
            local_run(
                arguments,
                {
                    'GRIZZLY_TEST_VAR': 'True',
                },
                {
                    'master': ['--foo', 'bar', '--master'],
                    'worker': ['--bar', 'foo', '--worker'],
                    'common': ['--common', 'true'],
                },
            )
            == 0
        )

        assert run_command.call_count == 1
        args, _ = run_command.call_args_list[-1]
        assert args[0] == [
            'behave',
            f'{test_context}/test.feature',
            '--foo',
            'bar',
            '--master',
            '--bar',
            'foo',
            '--worker',
            '--common',
            'true',
        ]

        assert environ.get('GRIZZLY_TEST_VAR', None) == 'True'

        assert (
            local_run(
                arguments,
                {
                    'GRIZZLY_TEST_VAR': 'True',
                },
                {
                    'master': ['--foo', 'bar', '--master'],
                    'worker': ['--bar', 'foo', '--worker'],
                    'common': ['--common', 'true', '-Dcsv-prefix="cool beans"'],
                },
            )
            == 0
        )

        assert run_command.call_count == 2
        args, _ = run_command.call_args_list[-1]
        assert args[0] == [
            'behave',
            f'{test_context}/test.feature',
            '--foo',
            'bar',
            '--master',
            '--bar',
            'foo',
            '--worker',
            '--common',
            'true',
            '-Dcsv-prefix="cool beans"',
        ]

        assert environ.get('GRIZZLY_TEST_VAR', None) == 'True'
    finally:
        rm_rf(test_context)
        with suppress(KeyError):
            del environ['GRIZZLY_TEST_VAR']
